#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <errno.h>
#include <stdarg.h>
#include "iolib.h"

/* The granularity with which to allocate buffer sizes;
   the allocated size of a buffer will always be a multiple of this. */
#define BUFFERUNIT 256

/* Default buffer size for iopenrb and iopenwb. */
#define DEFAULTBUFFERSIZE 4096

/* Default buffer size for breada. */
#define DEFAULTBLOCKSIZE 4096

/* Die with a constant error message. */
#define die(s) do { write(2, s, sizeof(s) - 1); exit(20); } while (0)

void bsetsize(buffer *b, int len) {
	if (len == 0) {
		if (b->s != NULL) free(b->s);
		b->s = NULL;
		b->len = b->size = 0;
	} else {
		/* This always rounds up the allocated size so that
		   adding characters to a buf one at a time doesn't
		   cause a realloc for each character. */
		int size = BUFFERUNIT * ((len / BUFFERUNIT) + 1);
		if (size != b->size) {
			b->s = realloc(b->s, size + 1);
			if (!b->s) die("out of memory");
			b->size = size;
		}
		b->len = len;
	}
}

void bfree(buffer *b) {
	bsetsize(b, 0);
}

void bmake(buffer *b, const char *s) {
	bsetsize(b, 0);
	bappendm(b, s, strlen(s));
}

void bappendm(buffer *dest, const char *src, int len) {
	int l = dest->len;
	bsetsize(dest, l + len);
	memcpy(dest->s + l, src, len);
}

void bappends(buffer *dest, const char *s) {
	bappendm(dest, s, strlen(s));
}

void bappendc(buffer *dest, char c) {
	bappendm(dest, &c, 1);
}

void bappend(buffer *dest, const buffer *src) {
	bappendm(dest, src->s, src->len);
}

int blength(const buffer *b) {
	return b->len;
}

const char *bstr(const buffer *b) {
	if (b->len == 0) return "";
	b->s[b->len] = '\0';
	return b->s;
}

void btrimr(buffer *b) {
	if (b->len == 0) return;
	while (b->len > 0) {
		char c = b->s[--b->len];
		if (!(c == ' ' || c == '\r' || c == '\n' || c == '\t')) break;
	}
	b->len++;
}

void breplacec(buffer *b, char from, char to) {
	char *p = b->s, *end = p + b->len;
	for (; p < end; p++) {
		if (*p == from) *p = to;
	}
}

int bwritea(const buffer *b, io *io) {
	return iwritea(io, b->s, b->len);
}

int bread(buffer *b, io *io, int n) {
	char *buf = alloca(n);
	int c;
	if (!buf) die("out of memory");
	c = iread(io, buf, n);
	if (c < 0) return -1;
	if (c == 0) return 0;
	bappendm(b, buf, c);
	return c;
}

int breada(buffer *b, io *io) {
	int c;
	do { c = bread(b, io, DEFAULTBLOCKSIZE); } while (c > 0);
	if (c < 0) return -1;
	return 0;
}

int breadterm(buffer *b, io *io, char term) {
	while (1) {
		char c;
		int r = iread(io, &c, 1);
		if (r < 0) return -1;
		if (r == 0) return 0;
		if (c == term) return 1;
		bappendc(b, c);
	}
}

int breadline(buffer *b, io *io) {
	return breadterm(b, io, '\n');
}

void istdinit() {
	iin = ifromfd(0);
	ibin = irbuf(iin, DEFAULTBUFFERSIZE, 0);
	iout = ifromfd(1);
	ibout = iwbuf(iout, DEFAULTBUFFERSIZE, 0);
	ierr = ifromfd(2);
	iberr = iwbuf(ierr, DEFAULTBUFFERSIZE, 0);
}

io *iopenr(const char *name) {
	return ifromfd(open(name, O_RDONLY));
}

io *iopenrb(const char *name) {
	return irbuf(iopenr(name), DEFAULTBUFFERSIZE, 1);
}

io *iopenrw(const char *name) {
	return ifromfd(open(name, O_RDWR));
}

io *iopenw(const char *name) {
	return ifromfd(open(name, O_WRONLY | O_TRUNC | O_CREAT, 0644));
}

io *iopenwb(const char *name) {
	return iwbuf(iopenw(name), DEFAULTBUFFERSIZE, 1);
}

io *iopenwexcl(const char *name) {
	return ifromfd(open(name, O_WRONLY | O_CREAT | O_EXCL, 0644));
}

io *iopenwa(const char *name) {
	return ifromfd(open(name, O_WRONLY | O_APPEND | O_CREAT, 0644));
}

io *ifromfd(int fd) {
	io *io;
	if (fd < 0) return NULL;
	io = malloc(sizeof *io);
	if (!io) die("out of memory");
	io->type = IOTYPE_FD;
	io->u.fd = fd;
	return io;
}

io *ifromsock(int sock) {
	io *io;
	if (sock < 0) return NULL;
	io = malloc(sizeof *io);
	if (!io) die("out of memory");
	io->type = IOTYPE_SOCK;
	io->u.sock = sock;
	return io;
}

io *irbuf(io *from, int bufsize, int close) {
	char *buf = malloc(bufsize);
	io *io;
	if (!from) return NULL;
	if (!buf) die("out of memory");
	io = malloc(sizeof *io);
	if (!io) die("out of memory");
	io->type = IOTYPE_RBUF;
	io->u.b.from = from;
	io->u.b.buf = buf;
	io->u.b.len = bufsize;
	io->u.b.start = io->u.b.end = 0;
	io->u.b.close = close;
	return io;
}

io *iwbuf(io *from, int bufsize, int close) {
	/* wbufs are the same as rbufs */
	io *io = irbuf(from, bufsize, close);
	if (!io) return NULL;
	io->type = IOTYPE_WBUF;
	return io;
}

int iflush(io *io) {
	switch (io->type) {
	case IOTYPE_WBUF:
		if (io->u.b.end > 0) {
			if (iwritea(io->u.b.from, io->u.b.buf, io->u.b.end) < 0) return -1;
			io->u.b.end = 0;
		}
		return 0;
	case IOTYPE_RBUF:
		io->u.b.start = io->u.b.end = 0;
		return 0;
	default:
		return 0;
	}
}

int iclose(io *io) {
	if (iflush(io) < 0) return -1;

	switch (io->type) {
	case IOTYPE_FD:
		close(io->u.fd);
		break;
	case IOTYPE_SOCK:
		shutdown(io->u.sock, 2);
		break;
	case IOTYPE_WBUF:
	case IOTYPE_RBUF:
		if (io->u.b.close) {
			if (iclose(io->u.b.from) < 0) return -1;
		}
		free(io->u.b.buf);
		break;
	}

	free(io);
	return 0;
}

int iread(io *io, char *buf, int n) {
	int c;
	switch (io->type) {
	case IOTYPE_FD:
		do {
			c = read(io->u.fd, buf, n);
		} while (c < 0 && errno == EINTR);
		break;
	case IOTYPE_SOCK:
		do {
			c = recv(io->u.sock, buf, n, 0);
		} while (c < 0 && errno == EINTR);
		break;
	case IOTYPE_RBUF:
		c = 0;
		while (n > 0) {
			int amt = io->u.b.end - io->u.b.start;
			if (amt == 0) {
				/* Buffer empty, read into it */
				io->u.b.start = 0;
				amt = io->u.b.end = iread(io->u.b.from,
					io->u.b.buf, io->u.b.len);
				if (amt <= 0) {
					c = amt;
					break;
				}
			} else {
				amt = amt < n ? amt : n;
				memcpy(buf, io->u.b.buf + io->u.b.start,
					amt);
				buf += amt;
				c += amt;
				n -= amt;
				io->u.b.start += amt;
			}
		}
		break;
	default:
		c = -1;
	}

	return c;
}

int iwrite(io *io, const char *buf, int n) {
	int c;

	switch (io->type) {
	case IOTYPE_FD:
		do {
			c = write(io->u.fd, buf, n);
		} while (c < 0 && errno == EINTR);
		break;
	case IOTYPE_SOCK:
		do {
			c = send(io->u.sock, buf, n, 0);
		} while (c < 0 && errno == EINTR);
		break;
	case IOTYPE_WBUF:
		c = n;
		while (n > 0) {
			int space = io->u.b.len - io->u.b.end;
			if (space == 0) {
				if (iflush(io) < 0) return -1;
			} else {
				space = space < n ? space : n;
				memcpy(io->u.b.buf + io->u.b.end,
					buf, space);
				buf += space;
				n -= space;
				io->u.b.end += space;
			}
		}
		break;
	default:
		c = -1;
	}

	return c;
}

int iwritea(io *io, const char *buf, int n) {
	int c;
	while (n > 0) {
		c = iwrite(io, buf, n);
		if (c < 0) return -1;
		buf += c;
		n -= c;
	}
	return 0;
}

int isync(io *io) {
	if (fsync(io->u.fd) < 0)
		return -1;
	return 0;
}

/* Return the representation of a digit. */
static char format_digit(int n) {
	if (n < 10) return '0' + n;
	if (n < 36) return 'A' + (n - 10);
	return '!';
}

/* Format an unsigned long into the buffer. */
static void format_ulong(buffer *b, unsigned long n, int base) {
	if (n >= base) format_ulong(b, n / base, base);
	bappendc(b, format_digit(n % base));
}

/* Format a signed long into the buffer. */
static void format_long(buffer *b, long n, int base) {
	if (n < 0) {
		bappendc(b, '-');
		format_ulong(b, -n, base);
	} else {
		format_ulong(b, n, base);
	}
}

/* Format an argument into the buffer. */
static void format_arg(buffer *b, char t, va_list *va) {
	switch (t) {
	/* char ('c') and buf ('b') are special-cased below for efficiency. */
	case 'i': {
		int i = va_arg(*va, int);
		format_long(b, i, 10);
		break;
	}
	case 'I': {
		unsigned int i = va_arg(*va, unsigned int);
		format_ulong(b, i, 10);
		break;
	}
	case 'x': {
		int i = va_arg(*va, int);
		format_long(b, i, 16);
		break;
	}
	case 'X': {
		unsigned int i = va_arg(*va, unsigned int);
		format_ulong(b, i, 16);
		break;
	}
	case 'l': {
		long i = va_arg(*va, long);
		format_long(b, i, 10);
		break;
	}
	case 'L': {
		unsigned long i = va_arg(*va, unsigned long);
		format_ulong(b, i, 10);
		break;
	}
	case 'a': {
		unsigned long i = va_arg(*va, unsigned long);
		format_long(b, i & 0xff, 10);
		bappendc(b, '.');
		format_long(b, (i >> 8) & 0xff, 10);
		bappendc(b, '.');
		format_long(b, (i >> 16) & 0xff, 10);
		bappendc(b, '.');
		format_long(b, (i >> 24) & 0xff, 10);
		break;
	}
	default: 
		die("unknown format character");
	}
}

int iformat(io *io, ...) {
	int r = 0;
	va_list va;
	va_start(va, io);
	while (1) {
		/* char gets promoted to int in varargs calls */
		char t = va_arg(va, int);
		if (t == 0) break;
		switch (t) {
		case 'c': {
			char *s = va_arg(va, char *);
			if (iwritea(io, s, strlen(s)) < 0) r = -1;
			break;
		}
		case 'b': {
			buffer *b = va_arg(va, buffer *);
			if (bwritea(b, io) < 0) r = -1;
			break;
		}
		default: {
			buffer b = BUFFER;
			format_arg(&b, t, &va);
			if (bwritea(&b, io) < 0) r = -1;
			bfree(&b);
		}
		}
	}
	va_end(va);
	return r;
}

void bformat(buffer *b, ...) {
	va_list va;
	va_start(va, b);
	while (1) {
		char t = va_arg(va, int);
		if (t == 0) break;
		switch (t) {
		case 'c': {
			char *s = va_arg(va, char *);
			bappends(b, s);
			break;
		}
		case 'b': {
			buffer *src = va_arg(va, buffer *);
			bappend(b, src);
			break;
		}
		default:
			format_arg(b, t, &va);
		}
	}
	va_end(va);
}

